package com.example;

import java.awt.image.BufferedImage;
import java.io.ByteArrayInputStream;
import java.util.ArrayList;
import java.util.Base64;
import java.util.concurrent.atomic.AtomicInteger;

import javax.imageio.ImageIO;

import org.json.JSONArray;
import org.json.JSONObject;

import com.example.types.Fingerprint;
import com.example.types.FingerprintCompareListItem;
import com.example.types.PersonFingerprintData;
import com.example.types.VerificationResult;
import static com.example.utils.JavaResponse.buildJavaResponse;
import com.example.utils.TransparencyContents;
import com.example.utils.Verification;
import com.machinezoo.sourceafis.FingerprintImage;
import com.machinezoo.sourceafis.FingerprintTemplate;

public class EventHandler {

    private static FingerprintCompareListItem[] registeredFingerprints = new FingerprintCompareListItem[0];

    public static String handleEvent(String rawMessage) {
        try {
            // System.err.println(rawMessage);  // Log raw message for debugging

            JSONObject message = new JSONObject(rawMessage);

            // Extract key fields from the JSON payload
            String id = message.optString("id", "unknown");
            String cmd = message.optString("cmd", "").toLowerCase(); // Normalize command
            String imageBase64 = message.optString("fpScan", null);  // Base64 fingerprint image
            int fingerPosition = message.optInt("fingerPosition", -1); // Optional finger position
            JSONArray templateArray = message.optJSONArray("fpTemplateArray"); // Existing templates
            JSONObject personInfoPayload = message.optJSONObject("personInfo"); // Optional person metadata This should be PersonFingerprintData

            PersonFingerprintData personInfo = new PersonFingerprintData(
                    personInfoPayload.optString("id"),
                    personInfoPayload.optString("name"),
                    personInfoPayload.optString("org"),
                    personInfoPayload.optString("code"),
                    new ArrayList<Fingerprint>()
            );

            FingerprintCompareListItem[] fpTemplateList = new FingerprintCompareListItem[templateArray.length()];
            for (int i = 0; i < templateArray.length(); i++) {
                JSONObject obj = templateArray.getJSONObject(i);
                String name = obj.optString("name");
                int fpPosition = obj.optInt("fpPosition");
                String fpTemplateString = obj.optString("fpTemplate");

                // Decoding it back to fingeprintTemplate so that afis can compare
                // byte[] templateBytes = Base64.getDecoder().decode(fpTemplateString);
                // FingerprintTemplate fpTemplate = new FingerprintTemplate(templateBytes);
                try {
                    byte[] templateBytes = Base64.getDecoder().decode(fpTemplateString);
                    System.out.println("Decoded bytes length: " + templateBytes.length);

                    FingerprintTemplate fpTemplate = new FingerprintTemplate(templateBytes);
                    System.out.println("Successfully created FingerprintTemplate!");
                    fpTemplateList[i] = new FingerprintCompareListItem(name, fpPosition, fpTemplate);
                } catch (Exception e) {
                    System.err.println("Error decoding or constructing FingerprintTemplate: " + e.getMessage());
                    e.printStackTrace();
                }

                // fpTemplateList[i] = new FingerprintCompareListItem(name, fpPosition, fpTemplate);
            };

            if (imageBase64 == null || imageBase64.isEmpty()) {
                throw new IllegalArgumentException("Missing fpScan (base64 image) data.");
            }

            // Route command to appropriate handler
            switch (cmd) {
                case "registration":
                    return handleRegistration(id, imageBase64, fpTemplateList, personInfo, fingerPosition);
                case "verification":
                    return handleVerification(id, imageBase64, fpTemplateList, personInfo, fingerPosition);
                case "qualityassurance":
                    return handleQualityAssurance(id, imageBase64, fpTemplateList, personInfo, fingerPosition);
                case "updateRegisteredFingerprints":
                    return updateRegisteredFingerprints(id, fpTemplateList);
                default:
                    // Unknown command fallback
                    JSONObject unknown = new JSONObject();
                    unknown.put("id", id);
                    unknown.put("message", "Unknown command: " + cmd);
                    return unknown.toString();
            }

        } catch (Exception e) {
            // Generic error response
            JSONObject error = new JSONObject();
            error.put("message", "Failed to process message: " + e.getMessage());
            return error.toString();
        }
    }

    public static String handleRegistration(String id, String imageBase64, FingerprintCompareListItem[] templateArray, PersonFingerprintData personInfo, int fingerPosition) {
        try {
            // Decode the fingerprint image from base64
            AtomicInteger edgeScore = new AtomicInteger();
            byte[] imageBytes = Base64.getDecoder().decode(imageBase64);
            BufferedImage image = ImageIO.read(new ByteArrayInputStream(imageBytes));
            // Create a FingerprintTemplate from the imageBytes
            FingerprintTemplate probeTemplate = new FingerprintTemplate(new FingerprintImage(imageBytes));

            if (image == null) {
                throw new IllegalArgumentException("Invalid image format.");
            }

            // Calculate edge scores to determine fingerprint quality
            edgeScore.set(TransparencyContents.calculateEdgeScore(imageBytes));

            // If no existing templates are provided, register the new one
            if (templateArray == null || templateArray.length <= 0 && registeredFingerprints.length <= 0) {
                // Register this new template
                String response = buildJavaResponse(id, "Registration", "Success", "No existing templates. Registered as new.", probeTemplate, 100.0, edgeScore.get(), personInfo, fingerPosition);
                return response.toString();  // Return the response as a JSON string
            }

            VerificationResult matchedPerson = Verification.verifyFingeprint(probeTemplate, templateArray, registeredFingerprints);
            // Return response based on score
            if (matchedPerson != null) { // Match found
                String response = buildJavaResponse(id, "Registration", "Registered", "Existing Fingerprint template found for " + matchedPerson + "!", probeTemplate, matchedPerson.getScore(), edgeScore.get(), personInfo, fingerPosition);
                return response;
            } else {
                // Register new template
                String response = buildJavaResponse(id, "Registration", "Success", "New fingerprint template constructed.", probeTemplate, 100.0, edgeScore.get(), personInfo, fingerPosition);
                return response.toString();
            }

        } catch (Exception e) {
            // Handle errors during registration
            String errorResponse = buildJavaResponse(id, "Registration", "Failed", "Template Construction Failed: " + e.getMessage(), null, 0.0, 0, personInfo, fingerPosition);
            return errorResponse.toString();
        }
    }

    public static String handleVerification(String id, String imageBase64, FingerprintCompareListItem[] templateArray, PersonFingerprintData personInfo, int fingerPosition) {
        try {
            // Decode the fingerprint image from base64
            AtomicInteger edgeScore = new AtomicInteger();
            byte[] imageBytes = Base64.getDecoder().decode(imageBase64);
            BufferedImage image = ImageIO.read(new ByteArrayInputStream(imageBytes));
            if (image == null) {
                throw new IllegalArgumentException("Invalid image format.");
            }

            // Create a FingerprintTemplate from the imageBytes
            FingerprintTemplate probeTemplate = new FingerprintTemplate(new FingerprintImage(imageBytes));

            // Calculate edge scores to determine fingerprint quality
            edgeScore.set(TransparencyContents.calculateEdgeScore(imageBytes));

            VerificationResult matchedPerson = Verification.verifyFingeprint(probeTemplate, templateArray, registeredFingerprints);

            // Return response based on score
            if (matchedPerson != null) { // Match found
                return buildJavaResponse(id, "Verification", "Registered", "Existing Fingerprint template found for " + matchedPerson.getName() + "!", probeTemplate, matchedPerson.getScore(), edgeScore.get(), personInfo, fingerPosition).toString().toString();
            } else {
                // Register new template
                return buildJavaResponse(id, "Verification", "Not Registered", "Fingerprint Data Not found.", probeTemplate, 0, 0, personInfo, fingerPosition);
            }

        } catch (Exception e) {
            String errorResponse = buildJavaResponse(id, "Verification", "Failed", "Verification operation faulty...", null, 0, 0, personInfo, fingerPosition); // defaul value
            return errorResponse.toString();
        }
    }

    public static String handleQualityAssurance(String id, String imageBase64, FingerprintCompareListItem[] templateArray, PersonFingerprintData personInfo, int fingerPosition) {
        try {
            AtomicInteger edgeScore = new AtomicInteger();
            // Decode the fingerprint image from base64
            byte[] imageBytes = Base64.getDecoder().decode(imageBase64);
            BufferedImage image = ImageIO.read(new ByteArrayInputStream(imageBytes));
            if (image == null) {
                throw new IllegalArgumentException("Invalid image format.");
            }

            // Create a FingerprintTemplate from the imageBytes
            FingerprintTemplate probeTemplate = new FingerprintTemplate(new FingerprintImage(imageBytes));

            // Calculate edge scores to determine fingerprint quality
            edgeScore.set(TransparencyContents.calculateEdgeScore(imageBytes));
            System.out.println("Edge Score " + edgeScore);

            VerificationResult matchedPerson = Verification.verifyFingeprint(probeTemplate, templateArray, registeredFingerprints);

            // Return response based on score
            if (matchedPerson != null) { // Match found
                return buildJavaResponse(id, "QualityAssurance", "Registered", "Registered fingeprint. ", null, 0, edgeScore.get(), personInfo, fingerPosition).toString();
            } else {
                return buildJavaResponse(id, "QualityAssurance", "Not Registered", "New Fingerprint detected...", probeTemplate, 100, edgeScore.get(), personInfo, fingerPosition).toString();
            }

        } catch (Exception e) {
            // Handle errors during checking
            String errorResponse = buildJavaResponse(id, "QualityAssurance", "Failed", "Unable to produce Edge score... " + e.getMessage(), null, 0.0, 0, personInfo, fingerPosition);
            return errorResponse.toString();
        }
    }

    public static String updateRegisteredFingerprints(String id, FingerprintCompareListItem[] fpTemplateList) {
        try {
            return "response here for now...";
        } catch (Exception e) {
            return "error response here for now...";
        }
    }

}
